
import re
import os
import sys
import json
import psycopg2
import schema
import argparse as ap
import cx_Oracle as almadb

from _version import ___version___ as version
from datetime import datetime
from pycapo import CapoConfig
from pathlib import Path
from . import wf, WorkflowTracker

# Alma Restores CLI:
_RUN_ALMA_RESTORE_DESCRIPTION = """AAT/PPI ALMA Data Restore, version {}:  This tool takes a Member OUS Status UID, 
 and from there it launches an ALMA specific version of the Restore workflow using the current version of the 
 calibration information from the NAASC archives. """



def _make_alma_restore_parser():
    r"""Command-line ALMA Restore Launch Parser"""
    result = ap.ArgumentParser(description=_RUN_ALMA_RESTORE_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. nmtest, mnprod')
    result.add_argument('-C', '--casa', action='store', default='',
                        help='Path to Casa package to use')
    result.add_argument('-E', '--email', action='store', default=None,
                        help='Optional email address for a completion message')
    # Placeholder, workflow can't do this yet.
    # result.add_argument('-D', '--destination', action='store', default='', help='Override location for CMS delivery')
    result.add_argument('mousUid', type = str, action='store', help='Member ObsUnitSet Identifier to be Restored')

    return result


def connect_alma(config):
    return pyat.schema.create_session('ALMA').connection().engine.raw_connection()

def connect_metadata(config):
    return pyat.schema.create_session('SDM').connection().engine.raw_connection()


def lookup_project_code(mous_uid):
    r"""We want the project code to simplify the job of the workflow (and, as a smoke test for whether
    the alma archive will behave itself for this process)"""

    configuration = CapoConfig()

    alma_handle = connect_alma(configuration).cursor()

    _ALMA_PC_QUERY="select ASA_PROJECT_CODE from ASA_OUS where ASA_OUS_ID= :mous_id"

    alma_handle.execute(_ALMA_PC_QUERY, mous_id=mous_uid)

    result = alma_handle.fetchone()[0]

    return result


def lookup_single_locator(mous_uid):
    configuration = CapoConfig()

    _AAT_PL_QUERY = """SELECT science_product_locator FROM execution_blocks WHERE alma_ous_id=%s"""

    aat_handle = connect_metadata(configuration).cursor()
    aat_handle.execute(_AAT_PL_QUERY, [mous_uid])

    return aat_handle.fetchone()[0]


def lookup_eb_uid(mous_uid):
    r"""Another query, we need a fileSetId related to our MOUS"""
    configuration = CapoConfig()

    alma_handle = connect_alma(configuration).cursor()

    _ALMA_EB_QUERY="select ASDM_UID from ASA_SCIENCE where MEMBER_OUSS_ID= :mous_id"

    alma_handle.execute(_ALMA_EB_QUERY, mous_id=mous_uid)

    # We'll get a bunch of results, but we only need one ASDM to proceed
    result = alma_handle.fetchone()[0]

    return result


def send_alma_ous_workflow_start(args, project_code, locator, file_set_id, tracking_key, cube_params=None):
    r"""Build the arguments for the WF call, and off we go!"""
    # An Example Command Line from workflow testing:
    #
    # wf -P nmtest -s DSOC
    # -p ousStatusId=uid://A001/X1284/X265f
    # -p fileSetIds=uid://A002/Xc7111c/X3979
    # -p telescope=ALMA
    # -p projectCodeOrDataType=2017.1.00886.L
    # runAlmaRestoreTest
    #
    # Build the wf command line, starting with all the fixed pieces:
    arguments = "-s NAASC -p processingSite=NAASC -p telescope=ALMA"
    # Required Parameters
    arguments += " -P " + os.environ['CAPO_PROFILE']
    arguments += " -p productLocator=" + locator
    arguments += " -p fileSetIds=" + file_set_id
    arguments += " -p ousStatusId=" + args.mousUid
    arguments += " -p projectCodeOrDataType=" + project_code
    arguments += " -p cliCorrId=" + tracking_key

    #     # Initial attempt to get it working again:  Give it an explict working directory:
    #     santized_mous = args.mousUid.replace(':', '_').replace('/', '_')
    #     timestamp = datetime.now().isoformat().replace('-','_').replace(':','_')
    #     arguments += " -p relativePath="+santized_mous+'_'+timestamp

    # Control whether we just restore, or image as well:
    if cube_params is None:
        arguments += " -p workflowName=AlmaOusRestoreWorkflow"
        arguments += " -p downloadDataFormat=CMS"
        arguments += " -p deliveryFormat=CMS"
    else:
        arguments += " -p workflowName=AlmaOusImagingWorkflow"
        arguments += " -p downloadDataFormat=IMG"
        arguments += " -p deliveryFormat=IMG"
        arguments += " -p optimizedImagingParameters=" + cube_params
    # The Optional Parameters:
    if '' != args.casa:
        arguments += " -p casaHome=" + args.casa
    else:
        config = CapoConfig()
        arguments += " -p casaHome=" + config['edu.nrao.archive.workflow.config.CasaVersions.homeForAlmaRestore']
    # Option not available yet, so don't display or use it
    # if '' != args.destination:
    #     arguments += " -p deliveryDirectory=" + args.destination
    # The Finale
    if cube_params is None:
        arguments += " runAlmaBasicRestoreWorkflow"
    else:
        arguments += " runAlmaOptimizedImagingWorkflow"
    # Build our command line and execute:
    wf(args=arguments.split(' '))

    # Talk to the user, these CLIs are bad on feedback
    print("Initiating processing of {} to spool.\n".format(args.mousUid))


def launch_alma_restore():
    r"""Take given mous_uid and launch a restore for it.  One/command, these are expensive."""

    # parse our arguments, get profile and list of Ids
    args = _make_alma_restore_parser().parse_args()

    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]

        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:

        # In order to use the direct reply functionality, we need to
        # send these messages via the NM system.
        if "va" in args.profile:
            profile_to_use = args.profile.replace('va', 'nm')
        else:
            profile_to_use = args.profile

        os.environ['CAPO_PROFILE'] = profile_to_use

    # Obtain the Project Code & an ASDM UID:  Require queries of the ALMA database:
    project_code = lookup_project_code(args.mousUid)
    asdm_uid = lookup_eb_uid(args.mousUid)
    asdm_locator = lookup_single_locator(args.mousUid)

    # Send an event to initiate workflow monitoring:

    my_tracker = WorkflowTracker(args.profile)
    monitor_key = my_tracker.prepareWorkflowTracking("runAlmaRestoreWorkflow", email=args.email)

    # Launch the workflow
    send_alma_ous_workflow_start(args, project_code, asdm_locator, asdm_uid, monitor_key)

    # Now wait for information:
    my_tracker.waitForReply()

    # and tell the user where things are happening:
    print("Alma Restore ({}) running in: {}".format(my_tracker.getLoggingId(), my_tracker.getWorkingDirectory()))


# Optimized Imaging
_RUN_ALMA_IMAGING_DESCRIPTION = """AAT/PPI ALMA Optimized Imaging, version {}:  This tool takes a Member OUS Status UID  
 and json file of parameters.  It then proceeds to initate the Optimized Imaging variant of the ALMA OUS processing 
 workflow.  It makes use of the template_hifa_cubeimage.xml recipe provided by the pipeline. """


def _make_alma_cube_imaging_parser():
    r"""Command-line ALMA Restore Launch Parser"""
    result = ap.ArgumentParser(description=_RUN_ALMA_RESTORE_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. vatest, vaprod')
    result.add_argument('-C', '--casa', action='store', default='',
                        help='Path to Casa package to use')
    result.add_argument('-E', '--email', action='store', default=None,
                        help='Optional email address for a completion message')
    result.add_argument('file', action='store', type=str, help='Name of the parameters file to use for imaging.')
    result.add_argument('mousUid', type = str, action='store', help='Member ObsUnitSet Identifier to be Imaged')

    return result


def launch_alma_cube_imaging():
    r"""
        A wrapper to set up the metadata for and tracking of an Alma Optimized
        Imaging workflow (note: this is a variant on the restore above.  Except
        for the imaging parameters & the downloadDataFormat value, things should
        be pretty much the same.
    """

    args = _make_alma_cube_imaging_parser().parse_args()

    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]

        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:

        # In order to use the direct reply functionality, we need to
        # send these messages via the NM system.
        if "va" in args.profile:
            profile_to_use = args.profile.replace('va', 'nm')
        else:
            profile_to_use = args.profile

        os.environ['CAPO_PROFILE'] = profile_to_use

    parameters_file = Path(args.file)
    if not parameters_file.exists():
        print(f"Can't find file {parameters_file}! Not executing workflow.")
        exit(5)

    # if we got here, then the file exists, read the contents for passing on to the workflow:
    file_handle = open(parameters_file, 'r')
    raw_parameters = file_handle.read()
    # remove all tab/newline/spaces to allow easier writing of the json by peoples:
    parameters_json = re.sub(r"[\n\t\s]*", "", raw_parameters)

    try:
        structured_json = json.loads(parameters_json)
        print(f"Parameters loaded in: {structured_json}")
        # Note: the json parsing here is just for validation purposes.
    except ValueError:
        print(f"{parameters_file} is not properly formatted Json. Attempted to parse: {parameters_json}")
        exit(3)

    # Obtain the Project Code & an ASDM UID:  Require queries of the ALMA database:
    project_code = lookup_project_code(args.mousUid)
    asdm_uid = lookup_eb_uid(args.mousUid)
    asdm_locator = lookup_single_locator(args.mousUid)

    # Send an event to initiate workflow monitoring:
    my_tracker = WorkflowTracker(args.profile)
    monitor_key = my_tracker.prepareWorkflowTracking("runAlmaOptimizedCubeWorkflow", email=args.email)

    # Launch the workflow
    send_alma_ous_workflow_start(args, project_code, asdm_locator, asdm_uid, monitor_key, cube_params=parameters_json)

    # Now wait for information:
    my_tracker.waitForReply()

    # and tell the user where things are happening:
    print("Alma Cube Imaging ({}) running in: {}".format(my_tracker.getLoggingId(), my_tracker.getWorkingDirectory()))

# Optimized Imaging Reprocess

_RUN_AOI_REPRO_DESCRIPTION = """AAT/PPI Alma Imaging Reprocessing, version {}: This tool initiates a re-run of the CASA imaging pipeline on the desired directory.  The data are placed in the Spool for the processing and returned to the QA area upon successful completion."""


def _make_aoi_reprocess_parser():
    r"""Parser"""
    result = ap.ArgumentParser(description=_RUN_AOI_REPRO_DESCRIPTION.format(version),
                           formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                    help='profile name to use, e.g. test, production')
    result.add_argument('-I', '--image_casa', action='store', default='',
                        help='Specify a CASA for imaging purposes (overrides what was chosen in the UI)')
    result.add_argument('-K', '--restore_casa', action='store', default='',
                        help='The C5 Kludge(tm): Separates restore processing, and sets the CASA version for restore')
    result.add_argument('-r', '--request', action='store', default='', type=str,
                        help='The external RequestId for this run (if applicable)')
    result.add_argument('directoryOfInterest', type = str, action='store',
                        help='Name of the directories to be re-run through AOI')
    return result


def launch_reimaging_run():
    """From a (list of) directory(s) launch the AOI Reprocessing workflow for each"""

    #parse our arguments, get profile and list of Ids
    args = _make_aoi_reprocess_parser().parse_args()

    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile


    #
    # Command line we want is similar to that for CIPL reprocessing:
    # (processingSite for message routing)
    #
    # wf -P [nmtest|nmprod] -p workflowName=CiplReprocess -p processingSite=DSOC -p targetDirectory=arg runCiplReprocessingWorkflow
    #
    # Build up the arguments:
    arguments = "-p workflowName=AoiReprocess -p processingSite=NAASC -p telescope=ALMA -s NAASC"
    arguments += " -P " +os.environ['CAPO_PROFILE']
    arguments += " -p targetDirectory=" + args.directoryOfInterest

    if '' != args.image_casa:
        arguments += " -p casaHome=" + args.imate_casa

    # Setting this variable activates The C5 Kludge(tm)
    if '' != args.restore_casa:
        arguments += " -p restoreCasa=" + args.restore_casa

    if '' != args.request:
        arguments += " -p requestId=" + args.request

    arguments += " runAoiReprocessingWorkflow"
    #clFinish = " runTestWorkflow"

    wf(args=arguments.split(' '))
    print(f"Re-running CASA for {args.directoryOfInterest} in spool.  You'll recieve an email shortly.")

